# copyright ################################# #
# This file is part of the Xfields Package.   #
# Copyright (c) CERN, 2021.                   #
# ########################################### #

from pathlib import Path
_pkg_root = Path(__file__).parent.absolute()
